import os
import json
import random
import hashlib
from json import JSONDecodeError

from mona_sdk.client_exceptions import MonaInitializationException

NORMALIZED_HASH_DECIMAL_DIGITS = 7
NORMALIZED_HASH_PRECISION = 10 ** NORMALIZED_HASH_DECIMAL_DIGITS


def get_boolean_value_for_env_var(env_var, default_value):
    return {"True": True, "true": True, "False": False, "false": False}.get(
        os.environ.get(env_var), default_value
    )


def get_dict_value_for_env_var(env_var, cast_values=None, default_value=None):
    """
    Expects a valid json string (or empty). cast_values allows to cast (and verify
    success of casting) all values to the given type.
    """
    value = os.environ.get(env_var)
    if not value:
        return default_value
    try:
        config = json.loads(value)
        if type(config) is not dict:
            raise MonaInitializationException(
                f'Env {env_var} isn\'t a valid json *Object*. Received: "{value}"'
            )
        if cast_values:
            for key in config:
                config[key] = cast_values(config[key])
        return config
    except JSONDecodeError:
        raise MonaInitializationException(
            f'Env {env_var} must be a valid json string. Received: "{value}"'
        )
    except ValueError:
        raise MonaInitializationException(
            f"Env {env_var} object values must be of type {cast_values}."
        )


def is_dict_contains_fields(message_event, required_fields):
    return all((field in message_event for field in required_fields))


def remove_items_by_value(data, value_to_remove):
    """
    Return a copy of the given dict after removing the items with value_to_remove as
    value.
    """
    return {key: value for key, value in data.items() if value != value_to_remove}


def _calculate_normalized_hash(context_id):
    """
    Calculate a normalized hash of a context id (a fraction between 0 to 1).
    """
    context_id_encoded = context_id.encode("ascii")
    # We chose sha224 here to have a different and independent hashing than sha1 which
    # is used on our end for similar purposes.
    return (
        int(hashlib.sha224(context_id_encoded).hexdigest(), base=16)
        % NORMALIZED_HASH_PRECISION
    ) / NORMALIZED_HASH_PRECISION


def keep_message_after_sampling(context_id, sampling_rate):
    if context_id:
        return _calculate_normalized_hash(context_id) <= sampling_rate
    else:
        # TODO(Nemo): Allow getting seed from the user for random.random().
        return random.random() <= sampling_rate



import ijson as ijson

from mona_sdk import MonaSingleMessage, Client

# Set the path of the JSON file and the num of objects be batch
PATH_TO_RAW_RATINGS = "/home/smadar/workspace/mona-sdk/mona_sdk/ratings.json"
NUMBER_OF_MESSAGES_IN_BATCH = 100000

# Set the client
sdk_client = Client(
    api_key="75425b87-26df-4fce-948e-4e64540d3d8e",
    secret="e77a979d-8362-4f4b-812b-5552d223e540",
    user_id="69d85e92-2894-4f7c-a1fb-55312711745f",
    raise_authentication_exceptions=True,
    raise_export_exceptions=True,
)

messages = []
number_of_messages_left_in_current_batch = NUMBER_OF_MESSAGES_IN_BATCH
number_of_batches = 0
total_number_of_messages_sent = 0
number_of_items_iterated = 0  # Where the script is in the file, how many messages
# it already went through (can be different from total_number_of_messages_sent when
# there are failures).


# Set the record keys
with open(PATH_TO_RAW_RATINGS, "r") as read_fd:
    for record in ijson.items(read_fd, "item"):
        messages.append(
            MonaSingleMessage(
                message=record,
                contextClass="ratings",
                contextId=record["user"] + "_" + record["movie"],
                exportTimestamp=record["timestamp"],
            )
        )
        number_of_messages_left_in_current_batch -= 1

        # If we processed the number we wanted in a batch, then it's time to send
        # this batch.
        if number_of_messages_left_in_current_batch == 0:
            number_of_items_iterated += NUMBER_OF_MESSAGES_IN_BATCH
            try:
                # sdk_client.export_batch(messages)
                number_of_batches += 1
                total_number_of_messages_sent += NUMBER_OF_MESSAGES_IN_BATCH
                print(
                    f"Successfully sent messages: "
                    f"{number_of_items_iterated - NUMBER_OF_MESSAGES_IN_BATCH} to "
                    f"{number_of_items_iterated}"
                )
            except Exception:
                print(
                    f"Failed to send messages: "
                    f"{number_of_items_iterated - NUMBER_OF_MESSAGES_IN_BATCH} to "
                    f"{number_of_items_iterated}"
                )

            messages = []
            number_of_messages_left_in_current_batch = NUMBER_OF_MESSAGES_IN_BATCH

    # Here we process the leftovers that didn't reach the amount of the full batch.
    number_of_items_iterated += len(messages)
    try:
        # sdk_client.export_batch(messages)
        number_of_batches += 1
        total_number_of_messages_sent += len(messages)
        print(
            f"Sent msgs {number_of_items_iterated - len(messages)} to "
            f"{number_of_items_iterated}"
        )
    except Exception:
        print(
            f"Failed to send messages {number_of_items_iterated - len(messages)} to "
            f"{number_of_items_iterated}"
        )
    print(
        f"Sent {total_number_of_messages_sent} messages in {number_of_batches} batches."
    )



# ----------------------------------------------------------------------------
#    Copyright 2021 MonaLabs.io
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
# ----------------------------------------------------------------------------
"""
This module holds all authentication information and related functions. For a given
api_key, it can provide a new access token, refresh an expired access token or give
authentication status information.
"""
import os
import time
import datetime
from functools import wraps
from threading import Lock

import requests
from requests.models import Response

from .logger import get_logger
from .client_exceptions import MonaAuthenticationException

# A new token expires after 22 hours, REFRESH_TOKEN_SAFETY_MARGIN is the safety gap of
# time to refresh the token before it expires (i.e. - in case
# REFRESH_TOKEN_SAFETY_MARGIN = 2, and the token is about to expire in 2 hours or less,
# the client will automatically refresh the token to a new one).
REFRESH_TOKEN_SAFETY_MARGIN = datetime.timedelta(
    hours=int(os.environ.get("MONA_SDK_REFRESH_TOKEN_SAFETY_MARGIN", 12))
)

AUTH_API_TOKEN_URL = os.environ.get(
    "MONA_SDK_AUTH_API_TOKEN_URL",
    "https://monalabs.frontegg.com/identity/resources/auth/v1/api-token",
)
REFRESH_TOKEN_URL = os.environ.get(
    "MONA_SDK_REFRESH_TOKEN_URL",
    "https://monalabs.frontegg.com/identity/resources/auth/v1/api-token/"
    "token/refresh",
)
BASIC_HEADER = {"Content-Type": "application/json"}
TOKEN_EXPIRED_DATE_FORMAT = "%a, %d %b %Y %H:%M:%S GMT"


# This dict maps between every api_key (each api_key is saved only once in this dict)
# and its access token info (if the given api_key is authenticated it will contain the
# token itself, its expiration date and the key to refresh it, otherwise it will contain
# the errors that occurred while trying to authenticate).
API_KEYS_TO_TOKEN_DATA = {}

# Token data args names:
ERRORS = "errors"
EXPIRES = "expires"
ACCESS_TOKEN = "accessToken"
REFRESH_TOKEN = "refreshToken"
TIME_TO_REFRESH = "timeToRefresh"
IS_AUTHENTICATED = "isAuthenticated"

# TODO(anat): consider initializing a different lock for each api_key.
authentication_lock = Lock()


def first_authentication(mona_client):
    if not is_authenticated(mona_client.api_key):
        # Make sure only one instance of the client (with the given api_key) can get a
        # new token. That token will be shared between all instances that share an
        # api_key.
        with authentication_lock:
            # The inner check is needed to avoid multiple redundant authentications.
            if not is_authenticated(mona_client.api_key):
                response = _request_access_token_with_retries(mona_client)
                API_KEYS_TO_TOKEN_DATA[mona_client.api_key] = response.json()

                # response.ok will be True if authentication was successful and
                # false if not.
                _set_api_key_authentication_status(mona_client.api_key, response.ok)
                _calculate_and_set_time_to_refresh(mona_client.api_key)

    # If the authentication failed, handle error and return false.
    if not is_authenticated(mona_client.api_key):
        return _handle_authentications_error(
            f"Mona's client could not authenticate. "
            f"errors: {_get_error_string_from_token_info(mona_client.api_key)}",
            mona_client.raise_authentication_exceptions,
        )
    else:
        get_logger().info(
            f"New client token info: {API_KEYS_TO_TOKEN_DATA[mona_client.api_key]}"
        )
        return True


def _get_error_string_from_token_info(api_key):
    error_list = _get_token_info_by_api_key(api_key, ERRORS)
    return ", ".join(_get_token_info_by_api_key(api_key, ERRORS)) if error_list else ""


def _request_access_token_with_retries(mona_client):
    return _get_auth_response_with_retries(
        lambda: _request_access_token_once(mona_client.api_key, mona_client.secret),
        num_of_retries=mona_client.num_of_retries_for_authentication,
        auth_wait_time_sec=mona_client.wait_time_for_authentication_retries,
    )


def _request_refresh_token_with_retries(refresh_token_key, mona_client):
    return _get_auth_response_with_retries(
        lambda: _request_refresh_token_once(refresh_token_key),
        num_of_retries=mona_client.num_of_retries_for_authentication,
        auth_wait_time_sec=mona_client.wait_time_for_authentication_retries,
    )


def _get_auth_response_with_retries(
    response_generator,
    num_of_retries,
    auth_wait_time_sec,
):
    """
    Sends an authentication request (first time/refresh) with a retry mechanism.
    :param response_generator (lambda)
            A function call that sends the wanted REST request.
    :return: The response received from the authentication server.
    """
    for i in range(num_of_retries + 1):
        try:
            response = response_generator()
            # Check that response is json-serializable.
            response.json()
            # Got a response, log and break the retry loop.
            get_logger().info(f"Got an authentication response after {i} retries.")
            break

        except Exception:
            if i == num_of_retries:
                # Retried to authenticate num_of_retries times and failed due to
                # authentications server problems, return a response with the relevant
                # info.
                response = _create_a_bad_response(
                    '{"errors": ["Could not connect to authentication server",'
                    ' "Number of retries: ' + str(i) + '"]}'
                )
            else:
                # TODO(anat): support exponential growth in wait times between retries.
                # Has more retries, sleep before trying again.
                time.sleep(auth_wait_time_sec)

    return response


def _request_access_token_once(api_key, secret):
    """
    Sends an access token REST request and returns the response.
    """
    return requests.request(
        "POST",
        AUTH_API_TOKEN_URL,
        headers=BASIC_HEADER,
        json={"clientId": api_key, "secret": secret},
    )


def _request_refresh_token_once(refresh_token_key):
    """
    Sends a refresh token REST request and returns the response.
    """
    return requests.request(
        "POST",
        REFRESH_TOKEN_URL,
        headers=BASIC_HEADER,
        json={"refreshToken": refresh_token_key},
    )


def _create_a_bad_response(content):
    """
    :param: content (str)
            The content of the response.
    :return: A functioning bad REST response instance with the given content.
    """
    response = Response()
    response.status_code = 400
    if type(content) is str:
        # _content expect bytes.
        response._content = bytes(content, "utf8")

    return response


def get_current_token_by_api_key(api_key):
    """
    :return: The given api_key's current access token.
    """
    return _get_token_info_by_api_key(api_key, ACCESS_TOKEN)


def _get_token_info_by_api_key(api_key, token_data_arg):
    """
    Returns the value of the wanted data for the given api_key.
    Returns None if the api_key or the arg does not exist.
    """
    return API_KEYS_TO_TOKEN_DATA.get(api_key, {}).get(token_data_arg)


def is_authenticated(api_key):
    """
    :return: True if Mona's client holds a valid token and can communicate with Mona's
    servers (or can refresh the token in order to), False otherwise.
    """
    return _get_token_info_by_api_key(api_key, IS_AUTHENTICATED)


def _set_api_key_authentication_status(api_key, bool_value):
    """
    Sets the IS_AUTHENTICATED arg in the token data dict of the given api_key, this
    setter is only needed to spare redundant calls for authentication.
    """
    API_KEYS_TO_TOKEN_DATA[api_key][IS_AUTHENTICATED] = bool_value


def _calculate_and_set_time_to_refresh(api_key):
    """
    Calculates the time the access token needs to be refreshed and updates the relevant
    api_key token data.
    """
    if is_authenticated(api_key):
        token_expires = datetime.datetime.strptime(
            _get_token_info_by_api_key(api_key, EXPIRES), TOKEN_EXPIRED_DATE_FORMAT
        )
        # Set the found value in the clients token info.
        API_KEYS_TO_TOKEN_DATA[api_key][TIME_TO_REFRESH] = (
            token_expires - REFRESH_TOKEN_SAFETY_MARGIN
        )


def _handle_authentications_error(
    error_message, should_raise_exception, message_to_log=None
):
    """
    Logs an error and raises MonaAuthenticationException if
    RAISE_AUTHENTICATION_EXCEPTIONS is true, else returns false.
    """
    get_logger().error(error_message)
    if message_to_log:
        get_logger().error(f"Failed to send the following to mona: {message_to_log}")
    if should_raise_exception:
        raise MonaAuthenticationException(error_message)
    return False


def _should_refresh_token(api_key):
    """
    :return: True if the token has expired, or is about to expire in
    REFRESH_TOKEN_SAFETY_MARGIN hours or less, False otherwise.
    """
    return (
        _get_token_info_by_api_key(api_key, TIME_TO_REFRESH) < datetime.datetime.now()
    )


def _refresh_token(mona_client):
    """
    Gets a new token and sets the needed fields.
    """
    refresh_token_key = _get_token_info_by_api_key(mona_client.api_key, REFRESH_TOKEN)
    response = _request_refresh_token_with_retries(refresh_token_key, mona_client)

    if not response.ok:
        get_logger().warning(
            f"Failed to refresh the access token, trying to get a new one. "
            f"{response.text}"
        )
        response = _request_access_token_with_retries(mona_client)

    authentications_response_info = response.json()

    # The current client token info will not change if the response was bad, so that on
    # the next function call the client will try to refresh the token again.
    if response.ok:
        # Update the client's new token info.
        API_KEYS_TO_TOKEN_DATA[mona_client.api_key] = authentications_response_info
        _set_api_key_authentication_status(mona_client.api_key, True)
        _calculate_and_set_time_to_refresh(mona_client.api_key)

        get_logger().info(
            f"Refreshed access token, the new token info:"
            f" {API_KEYS_TO_TOKEN_DATA[mona_client.api_key]}"
        )
    return response


def get_basic_auth_header(api_key, with_auth):
    return (
        {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {get_current_token_by_api_key(api_key)}",
        }
        if with_auth
        else BASIC_HEADER
    )


class Decorators(object):
    @classmethod
    def refresh_token_if_needed(cls, decorated):
        """
        This decorator checks if the current client's access token is about to
        be expired/already expired, and if so, updates to a new one.
        """

        @wraps(decorated)
        def inner(*args, **kwargs):
            # args[0] is the current mona_client instance.
            mona_client = args[0]

            if not mona_client.should_use_authentication:
                return decorated(*args, **kwargs)

            # If len(args) < 1, the wrapped function does not have args to log (neither
            # messages nor config)
            should_log_args = len(args) > 1 and mona_client.should_log_failed_messages

            # message_to_log is the messages/config that should be logged in case of
            # an authentication failure.
            message_to_log = args[1] if should_log_args else None

            if not is_authenticated(mona_client.api_key):
                return _handle_authentications_error(
                    "Mona's client is not authenticated",
                    mona_client.raise_authentication_exceptions,
                    message_to_log,
                )

            if _should_refresh_token(mona_client.api_key):
                with authentication_lock:
                    # The inner check is needed to avoid double token refresh.
                    if _should_refresh_token(mona_client.api_key):
                        refresh_token_response = _refresh_token(mona_client)
                        if not refresh_token_response.ok:
                            # TODO(anat): Check if the current token is still valid to
                            #   call the function anyway.
                            return _handle_authentications_error(
                                f"Could not refresh token: "
                                f"{refresh_token_response.text}",
                                mona_client.raise_authentication_exceptions,
                                message_to_log,
                            )
            return decorated(*args, **kwargs)

        return inner



from random import randint

from mona_sdk import Client, MonaSingleMessage

# Es monitor test
client_id = "e8018061-0559-4b22-b041-f10ef5848cd6"
secret_key = "9df5eefb-de6d-4529-b79a-60b398353449"

CONTEXT_CLASSES = [
    "SMADAR_TEST_IN_CONFIG_SAMPLED",
    "SMADAR_TEST_IN_CONFIG_SAMPLED_NOT_TAKEN",
    "SMADAR_TEST_TAKES_DEFAULT",
    "SMADAR_TEST_IN_CONFIG_FULLY_TAKEN",
]
NUMBER_OF_MESSAGES_PER_CONTEXT_CLASS = 10000


def random_with_N_digits(n):
    range_start = 10 ** (n - 1)
    range_end = (10 ** n) - 1
    return randint(range_start, range_end)


# Let's have 4 context classes - the 4th will take the default value (0.7).
my_mona_client = Client(
    api_key=client_id,
    secret=secret_key,
    user_id="ef3a7c1a-cf94-489b-8c69-5075136003f2",
    # should_use_authentication=False,
    # should_use_ssl=False,
    # override_rest_api_host="rest-api host address",
    sampling_config_name="smadar_neti_test"
)


for context_class in CONTEXT_CLASSES:
    messages = []
    for i in range(NUMBER_OF_MESSAGES_PER_CONTEXT_CLASS):
        messages.append(
            MonaSingleMessage(
                message={"sampled": "true"},
                contextClass=context_class,
                contextId=str(random_with_N_digits(10)),
            )
        )

    succeed_to_export = my_mona_client.export_batch(messages)
    print(f"\n{context_class}: {succeed_to_export}")



from mona_sdk import Client
# For regular smadar's user
client_id = "40c08ee6-f411-41b1-9752-80d98897aae6"
secret_key = "5694fad0-e372-488e-89cf-22ed8693a5a2"


my_mona_client = Client(
    api_key=client_id,
    secret=secret_key,
    user_id="aec02df0-813d-4f6a-8574-bdc2668c3cbc",
    #should_use_authentication=False,
    #should_use_ssl=False,
    #override_rest_api_host="rest-api host address",
    #context_class_to_sampling_rate={"SMADAR_TEST1": 0.5, "SMADAR_TEST2": 0.2},
    #default_sampling_rate=0.7
)

# print("------------------ get config ------------------")
# config = my_mona_client.get_config()
# print(config)

# print("------------------ get_suggested_config ------------------")
# res = my_mona_client.get_suggested_config()
# my_suggested_mona_config = my_mona_client.get_suggested_config()
# print(my_suggested_mona_config)
from mona_sdk import MonaSingleMessage, Client

config = {"user_id": 123}

# print("------------------ upload config ------------------")
# res = my_mona_client.upload_config(config, f"Anat test {time.time()}", "anat_local_test")
# print(res)


# print("------------------ send single message ------------------")
# example_to_send = {"ANAT_TEST": "b", "c": None, "test": 2, "a": False, "b": None}
# succeed_to_export = my_mona_client.export(MonaSingleMessage(
#     message=example_to_send,
#     contextClass='ANAT_TEST',
#     contextId='603c97a218d6812b4cce32fa1',
#     exportTimestamp="2021-05-04T21:55:23+00:00",
# ), filter_none_fields=False)
# print(succeed_to_export)


print("------------------ send batch ------------------")

#example_to_send = {"SMADAR_TEST": "a"}
succeed_to_export = my_mona_client.export_batch([
    MonaSingleMessage(
        message={"smadar_test": "a - should be added"},
        contextClass="SMADAR_TEST1",
        contextId='603c97a218d6812b4cce32fa',
    ),
    MonaSingleMessage(
        message={"smadar_test": "b - shouldn't be added 1"},
        contextClass="SMADAR_TEST2",
        contextId='903c97aagfd6812b4c8e32fa',
    ),
    MonaSingleMessage(
        message={"smadar_test": "c - shouldn't be added 2"},
        contextClass="SMADAR_TEST1",
        contextId='903c97aagfd6812b48gj27df',
    ),
    MonaSingleMessage(
        message={"smadar_test": "d - should be added or not, without context_id"},
        contextClass="SMADAR_TEST1",
    ),
    MonaSingleMessage(
        message={"smadar_test": "e - should be added or not, without context_id"},
        contextClass="SMADAR_TEST_NOT_IN_SAMPLING_CONFIG",
    ),
    MonaSingleMessage(
        message={"smadar_test": "f - should be added, with context_id"},
        contextClass="SMADAR_TEST_NOT_IN_SAMPLING_CONFIG",
        contextId='10jjf8e4gfd681204j38f7gb'
    )
])

print(succeed_to_export)
#
#
# a = {
#     "userId": "JMETER_TEST_USER_ID",
#     "messages": [
#         {
#             "arcClass": "JMETER_TEST",
#             "message": {
#                 "field_number_0": "we want each message to be 2k with 20 fields, here's a random numberrandom.nextInt()",
#                 "field_number_1": "we want each message to be 2k with 20 fields, here's a random numberrandom.nextInt()"
#             },
#             "contextId": "context_random.nextInt()",
#             "exportTimestamp": "1649215835"},
#         {
#             "arcClass": "JMETER_TEST",
#             "message": {
#                 "field_number_0": "we want each message to be 2k with 20 fields, here's a random numberrandom.nextInt()",
#                 "field_number_1": "we want each message to be 2k with 20 fields, here's a random numberrandom.nextInt()"
#             },
#             "contextId": "context_random.nextInt()",
#             "exportTimestamp": "1649215835"
#         }
#     ]
# }






# ----------------------------------------------------------------------------
#    Copyright 2021 MonaLabs.io
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
# ----------------------------------------------------------------------------
import os
import logging
from json import JSONDecodeError
from typing import List
from dataclasses import dataclass
from cachetools import TTLCache, cached

import jwt
import requests
from requests.exceptions import ConnectionError

from mona_sdk.client_exceptions import MonaServiceException, MonaInitializationException
from .logger import get_logger
from .validation import (
    handle_export_error,
    update_mona_fields_names,
    validate_inner_message_type,
    validate_mona_single_message,
    mona_messages_to_dicts_validation,
)
from .client_util import (
    remove_items_by_value,
    get_dict_value_for_env_var,
    keep_message_after_sampling,
    get_boolean_value_for_env_var,
)
from .authentication import (
    Decorators,
    is_authenticated,
    first_authentication,
    get_basic_auth_header,
    get_current_token_by_api_key,
)

# Note: if RAISE_AUTHENTICATION_EXCEPTIONS = False and the client could not
# authenticate, every function call will return false.
# Use client.is_active() in order to check authentication status.
RAISE_AUTHENTICATION_EXCEPTIONS = get_boolean_value_for_env_var(
    "MONA_SDK_RAISE_AUTHENTICATION_EXCEPTIONS", False
)

RAISE_EXPORT_EXCEPTIONS = get_boolean_value_for_env_var(
    "MONA_SDK_RAISE_EXPORT_EXCEPTIONS", False
)

RAISE_SERVICE_EXCEPTIONS = get_boolean_value_for_env_var(
    "MONA_SDK_RAISE_SERVICE_EXCEPTIONS", False
)

SHOULD_USE_AUTHENTICATION = get_boolean_value_for_env_var(
    "MONA_SDK_SHOULD_USE_AUTHENTICATION", True
)

SHOULD_USE_SSL = get_boolean_value_for_env_var("MONA_SDK_SHOULD_USE_SSL", True)

OVERRIDE_APP_SERVER_HOST = os.environ.get("MONA_SDK_OVERRIDE_APP_SERVER_HOST")

# TODO(anat): Once no one is using it, remove this env var (leave only
#  OVERRIDE_REST_API_HOST).
OVERRIDE_REST_API_URL = os.environ.get("MONA_SDK_OVERRIDE_REST_API_URL")
OVERRIDE_REST_API_HOST = os.environ.get("MONA_SDK_OVERRIDE_REST_API_HOST")

# Number of retries to authenticate in case the authentication server failed to
# respond.
NUM_OF_RETRIES_FOR_AUTHENTICATION = int(
    os.environ.get("MONA_SDK_NUM_OF_RETRIES_FOR_AUTHENTICATION", 3)
)

# Time to wait (in seconds) between retries in case the authentication server failed to
# respond.
WAIT_TIME_FOR_AUTHENTICATION_RETRIES_SEC = int(
    os.environ.get("MONA_SDK_WAIT_TIME_FOR_AUTHENTICATION_RETRIES_SEC", 2)
)

# When this variable is True, failed messages (for any reason) will be logged at "Error"
# level.
SHOULD_LOG_FAILED_MESSAGES = get_boolean_value_for_env_var(
    "MONA_SDK_SHOULD_LOG_FAILED_MESSAGES", False
)

FILTER_NONE_FIELDS_ON_EXPORT = get_boolean_value_for_env_var(
    "MONA_SDK_FILTER_NONE_FIELDS_ON_EXPORT", False
)

# SDK will randomly sample the sent data using this factor and disregard the sampled-
# out data, unless the sent data is set on a class overridden by
# MONA_SDK_SAMPLING_CONFIG.
DEFAULT_SAMPLING_FACTOR = float(os.environ.get("MONA_SDK_DEFAULT_SAMPLING_FACTOR", 1))

# When set, SDK will randomly sample the sent data for any class keyed in the config.
# See readme for more details.
SAMPLING_CONFIGURATION_DICT = get_dict_value_for_env_var(
    "MONA_SDK_SAMPLING_CONFIG", cast_values=float
)

SAMPLING_CONFIG_NAME = os.environ.get("SAMPLING_CONFIG_NAME")

SAMPLING_FACTORS_MAX_AGE_SECONDS = os.environ.get(
    "SAMPLING_FACTORS_MAX_AGE_SECONDS", 300
)

SERVICE_ERROR_MESSAGE = "Could not get server response for the wanted service."
UPLOAD_CONFIG_ERROR_MESSAGE = (
    "Could not upload the new configuration, please check it is valid."
)
APP_SERVER_CONNECTION_ERROR_MESSAGE = "Cannot connect to app-server."

UNAUTHENTICATED_ERROR_CHECK_MESSAGE = (
    "\nNotice that should_use_authentication is set to False, which is not supported by"
    " default and must be explicitly requested from Mona team."
)

# The argument to use as a default value on the values of the data argument (dict) when
# calling _app_server_request(). Use this and not None in order to be able to pass a
# None argument if needed.
UNPROVIDED_VALUE = "mona_unprovided_value"

# TODO(anat): change the following line once REST-api allows "contextClass"
#  instead of "arcClass".
CONTEXT_CLASS_FIELD_NAME = "arcClass"
CONTEXT_ID_FIELD_NAME = "contextId"


@dataclass
class MonaSingleMessage:
    """
    Class for keeping properties for a single Mona Message.
    Attributes:
        :param message (dict): (Required) JSON serializable dict with properties to send
            about the context ID.
        :param contextClass (str): (Required) context classes are defined in the Mona
            Config and define the schema of the contexts.
        :param contextId (str): (Optional) A unique identifier for the current context
            instance. One can export multiple messages with the same context_id and Mona
            would aggregate all of these messages to one big message on its backend.
            If none is given, Mona will create a random uuid for it. This is highly
            unrecommended - since it takes away the option to update this data in the
            future.
        :param exportTimestamp (int|str): (Optional) This is the primary timestamp Mona
            will use when considering the data being sent. It should be a date (ISO
            string or a Unix time number) representing the time the message was created.
            If not supplied, current time is used.
        :param action (str): (Optional) The action to use on the values in the fields of
            this message - "OVERWRITE", "ADD" or "NEW" (default: "OVERWRITE").


    A new message initialization would look like this:
    message_to_mona = MonaSingleMessage(
        message=<the relevant monitoring information>,
        contextClass="MY_CONTEXT_CLASS_NAME",
        contextId=<the context instance unique id>,
        exportTimestamp=<the message export timestamp>,
        action=<the wanted action>,
    )
    """

    message: dict
    contextClass: str
    contextId: str = None
    exportTimestamp: int or str = None
    action: str = None

    def get_dict(self):
        return {
            key: value
            for key, value in self.__dict__.items()
            if key in MonaSingleMessage.__dataclass_fields__.keys()
        }


class Client:
    """
    The main Mona python client class. Use this class to communicate with any of Mona's
    API.
    """

    def __init__(
        self,
        api_key=None,
        secret=None,
        raise_authentication_exceptions=RAISE_AUTHENTICATION_EXCEPTIONS,
        raise_export_exceptions=RAISE_EXPORT_EXCEPTIONS,
        raise_service_exceptions=RAISE_SERVICE_EXCEPTIONS,
        num_of_retries_for_authentication=NUM_OF_RETRIES_FOR_AUTHENTICATION,
        wait_time_for_authentication_retries=WAIT_TIME_FOR_AUTHENTICATION_RETRIES_SEC,
        should_log_failed_messages=SHOULD_LOG_FAILED_MESSAGES,
        should_use_ssl=SHOULD_USE_SSL,
        should_use_authentication=SHOULD_USE_AUTHENTICATION,
        override_rest_api_full_url=OVERRIDE_REST_API_URL,
        override_rest_api_host=OVERRIDE_REST_API_HOST,
        override_app_server_host=OVERRIDE_APP_SERVER_HOST,
        user_id=None,
        filter_none_fields_on_export=FILTER_NONE_FIELDS_ON_EXPORT,
        default_sampling_rate=DEFAULT_SAMPLING_FACTOR,
        context_class_to_sampling_rate=SAMPLING_CONFIGURATION_DICT,
        sampling_config_name=SAMPLING_CONFIG_NAME,
    ):
        """
        Creates the Client object. this client is lightweight so it can be regenerated
        or reused to user convenience.
        :param api_key: An api key provided to you by Mona.
        :param secret: The secret corresponding to the given api_key.
        """
        if not should_use_authentication and not user_id:
            raise MonaInitializationException(
                "When MONA_SDK_SHOULD_USE_AUTHENTICATION is turned off user_id must be "
                "provided."
            )

        if sampling_config_name and context_class_to_sampling_rate:
            raise MonaInitializationException(
                "Only one sampling method can be used at a time. Either remove "
                "sampling_config_name or context_class_to_sampling_rate"
            )

        self._logger = get_logger()

        self.api_key = api_key
        self.secret = secret

        self.raise_export_exceptions = raise_export_exceptions
        self.raise_service_exceptions = raise_service_exceptions
        self.should_log_failed_messages = should_log_failed_messages
        self.should_use_ssl = should_use_ssl
        self.should_use_authentication = should_use_authentication

        if should_use_authentication:
            self.raise_authentication_exceptions = raise_authentication_exceptions
            self.num_of_retries_for_authentication = num_of_retries_for_authentication
            self.wait_time_for_authentication_retries = (
                wait_time_for_authentication_retries
            )

            could_authenticate = first_authentication(self)
            if not could_authenticate:
                # TODO(anat): consider replacing this return with an if statement for
                #  the next part.
                return

        # If user_id=None then should_use_authentication must be True, which means at
        # this point the client was successfully authenticated and self._get_user_id()
        # will work.
        self._user_id = user_id or self._get_user_id()
        self._rest_api_url = (
            override_rest_api_full_url
            or self._get_rest_api_export_url(override_host=override_rest_api_host)
        )
        self._app_server_url = self._get_app_server_url(
            override_host=override_app_server_host
        )
        self.filter_none_fields_on_export = filter_none_fields_on_export

        # Data sampling.

        self._sampling_config_name = sampling_config_name
        self._context_class_to_sampling_rate = context_class_to_sampling_rate or {}
        self._default_sampling_rate = default_sampling_rate

        if self._sampling_config_name:
            sampling_factors_list = self.get_sampling_factors()

            if not sampling_factors_list:
                raise MonaInitializationException("Config name does not exist.")

            self._latest_seen_sampling_config = sampling_factors_list[0]

            self._context_class_to_sampling_rate = (
                self._latest_seen_sampling_config.get("factors_map", {})
            )
            self._default_sampling_rate = self._latest_seen_sampling_config.get(
                "default_factor", default_sampling_rate
            )

    def _get_rest_api_export_url(self, override_host=None):
        http_protocol = "https" if self.should_use_ssl else "http"
        host_name = override_host or f"incoming{self._user_id}.monalabs.io"
        endpoint_name = "export" if self.should_use_authentication else "monaExport"
        return f"{http_protocol}://{host_name}/{endpoint_name}"

    def _get_app_server_url(self, override_host=None):
        http_protocol = "https" if self.should_use_ssl else "http"
        host_name = override_host or f"api{self._user_id}.monalabs.io"
        return f"{http_protocol}://{host_name}"

    def is_active(self):
        """
        Returns True if the client is authenticated (or able to re-authenticate when
        needed) and therefore can export messages, otherwise returns False, meaning the
        client cannot export data to Mona's servers.
        Use this method to check client status in case RAISE_AUTHENTICATION_EXCEPTIONS
        is set to False.
        """
        return (
            is_authenticated(self.api_key) if self.should_use_authentication else True
        )

    def _get_user_id(self):
        """
        :return: The customer's user id (tenant id).
        """
        decoded_token = jwt.decode(
            get_current_token_by_api_key(self.api_key), verify=False
        )
        return decoded_token["tenantId"]

    @staticmethod
    def filter_none_fields(message):
        return {key: val for key, val in message.items() if val is not None}

    def should_filter_none_fields(self, filter_none_fields):
        """
        :param filter_none_fields:
            The value the export function got for filter_none_fields.
        :return: boolean
            True if the None fields should be filtered, if the caller function did
            not provide filter_none_fields use the client's self default.
        """
        return (
            self.filter_none_fields_on_export
            if filter_none_fields is None
            else filter_none_fields
        )

    @Decorators.refresh_token_if_needed
    def export(self, message: MonaSingleMessage, filter_none_fields=None):
        """
        Exports a single message to Mona's systems.

        :param message: MonaSingleMessage (required)
            message should be a MonaSingleMessage instance, which is a dataclass
            provided in this module.
        :param filter_none_fields: boolean (optional)
            When set to true fields with None values will be filtered out from the
            message dict.
        :return: boolean
            True if the message was successfully sent to Mona's systems,
            False otherwise (failure reason will be logged).
        """
        export_result = self._export_batch_inner(
            [message], filter_none_fields=filter_none_fields
        )
        return export_result and export_result["failed"] == 0

    @Decorators.refresh_token_if_needed
    def export_batch(
        self,
        events: List[MonaSingleMessage],
        default_action=None,
        filter_none_fields=None,
    ):
        """
        Use this function to easily send a batch of MonaSingleMessage to Mona.
        :param events: List[MonaSingleMessage] (required)
            Events should be a list of MonaSingleMessage (provided in this module).
        :param default_action: str (optional)
            The default action to the batch. Will be set as the action of all messages
            with no action provided.
        :param filter_none_fields: boolean (optional)
            When set to true fields with None values will be filtered out from the
            message dict.
        :return: dict
            Returns a dict of the following format:
            {
                "total": <number of messages event in batch>,
                "sent": <number of messages event successfully sent>,
                "failed": <number of messages event failed to be sent>
                "failure_reason": a dict of failed messages (index and contextId) and
                                  for each message the reason it failed.
            }
        """
        return self._export_batch_inner(
            events, default_action, filter_none_fields=filter_none_fields
        )

    def _should_add_message_to_sampled_data(self, message):
        context_class = message.get(CONTEXT_CLASS_FIELD_NAME)
        context_class_sampling_rate = self._context_class_to_sampling_rate.get(
            context_class
        )
        context_id = message.get(CONTEXT_ID_FIELD_NAME)
        if context_class_sampling_rate is not None:
            return keep_message_after_sampling(context_id, context_class_sampling_rate)
        return keep_message_after_sampling(context_id, self._default_sampling_rate)

    def _should_sample_data(self):
        return (self._default_sampling_rate < 1) or self._context_class_to_sampling_rate

    def _export_batch_inner(
        self,
        events: List[MonaSingleMessage],
        default_action=None,
        filter_none_fields=None,
    ):
        self._update_sampling_factors_if_needed()

        events = mona_messages_to_dicts_validation(
            events, self.raise_export_exceptions, self.should_log_failed_messages
        )
        if not events:
            return False

        messages_to_send = []
        for message_event in events:
            if not validate_mona_single_message(message_event):
                return handle_export_error(
                    "Messages to export must be of MonaSingleMessage type.",
                    self.raise_export_exceptions,
                    events if self.should_log_failed_messages else None,
                )

            message_copy = dict(message_event)

            # TODO(anat): remove the following line once REST-api allows "contextClass"
            #  instead of "arcClass".
            message_copy[CONTEXT_CLASS_FIELD_NAME] = message_copy.pop("contextClass")

            # TODO(anat): Add full validations on client side.
            if validate_inner_message_type(message_copy["message"]):
                # Change fields in message that starts with "MONA_".
                message_copy["message"] = update_mona_fields_names(
                    message_copy["message"]
                )

            if (
                self._should_sample_data()
                and not self._should_add_message_to_sampled_data(message_copy)
            ):
                logging.info(
                    f"This event isn't a part of the sampled data: {message_event}"
                )
                continue

            if self.should_filter_none_fields(filter_none_fields):
                message_copy["message"] = self.filter_none_fields(
                    message_copy["message"]
                )

            messages_to_send.append(message_copy)

        # Create and send the rest call to Mona's rest-api.
        try:
            rest_api_response = self._send_mona_rest_api_request(
                messages_to_send, default_action
            )
        except ConnectionError:
            return handle_export_error(
                "Cannot connect to rest-api",
                self.raise_export_exceptions,
                events if self.should_log_failed_messages else None,
            )

        # Create the response and return it.
        client_response = Client._create_client_response(
            rest_api_response,
            total=len(messages_to_send),
        )
        if client_response["failed"] > 0:
            handle_export_error(
                f"Some messages didn't pass validation: {client_response}."
                f"{self._get_unauthenticated_mode_error_message()}",
                self.raise_export_exceptions,
                events if self.should_log_failed_messages else None,
            )
        else:
            self._logger.info(
                f"All {client_response['total']} messages have been sent."
            )

        return client_response

    def _send_mona_rest_api_request(self, messages, default_action=None):
        """
        Sends a REST call to Mona's servers with the provided messages.
        :return: A REST response.
        """
        body = {"userId": self._user_id, "messages": messages}
        if default_action:
            body["defaultAction"] = default_action
        return requests.request(
            "POST",
            self._rest_api_url,
            headers=get_basic_auth_header(self.api_key, self.should_use_authentication),
            json=body,
        )

    @staticmethod
    def _create_client_response(
        response,
        total,
    ):
        """
        Creates the dict response of the client to an export_batch() call.
        :param response: The response received from the rest-api.
        :param total: total messages in batch.
        :return: A dict with response details (see export_batch docs for format).
        """
        # Sum the number of messaged that Mona's client didn't send to rest-api.
        failed = 0
        failure_reasons = {}

        # Check if some messages didn't passed validation on the rest-api.
        if not response.ok:
            try:
                result_info = response.json()
                failed = result_info["failed"]
                failure_reasons = result_info["failure_reasons"]
            except Exception:
                failed = total
                failure_reasons = "Failed to send the batch to Mona's servers"

        # Return the total result of the batch.
        return {
            "total": total,
            "failed": failed,
            "sent": total - failed,
            "failure_reasons": failure_reasons,
        }

    @Decorators.refresh_token_if_needed
    def upload_config(self, config, commit_message, author=None):
        """
        Uploads a new configuration, as a json-serializable dict.
        The configuration file enables you to define how the exported data should be
        aggregated and analyzed in Mona, as well as your insight and alerting
        preferences.
        Find more information on creating your configuration at:
        https://docs.monalabs.io/.
        :param config: (dict) Your new Mona configuration represented as a python dict
        (both the configuration dict with your user_id as the top key and just the
        configuration dict itself are accepted).
        :param commit_message: (str)
        :param author: (str) An email address identifying the configuration uploader.
        Mona will use this mail to send updates regarding re-creation of insights upon
        this configuration change. When not supplied, the author will be the Client's
        api-key and you will not get updates regarding the changes mentioned above.
        Must be provided when using un-authenticated mode.
        :return: A dict holding the upload data:
        {
            "success": <was the upload successful>, (bool)
            "new_config_id": <the new configuration ID> (str)
        }
        """
        upload_output = {"success": False, "new_config_id": ""}

        if not author and not self.should_use_authentication:
            self._handle_service_error(
                "When using non authenticated client, author must be provided."
            )
            return upload_output

        if not isinstance(config, dict):
            self._handle_service_error("config must be a dict.")
            return upload_output

        keys_list = list(config.keys())
        if len(keys_list) == 1 and keys_list[0] == self._user_id:
            config = config[keys_list[0]]

        config_to_upload = {
            "config": {self._user_id: config},
            "author": author or self.api_key,
            "commit_message": commit_message,
            "user_id": self._user_id,
        }

        try:
            upload_response = requests.post(
                f"{self._app_server_url}/upload_config",
                headers=get_basic_auth_header(
                    self.api_key, self.should_use_authentication
                ),
                json=config_to_upload,
            )
            response_data = upload_response.json()["response_data"]
            upload_output["new_config_id"] = response_data["new_config_id"]
            upload_output["success"] = upload_response.ok

            if not upload_output["success"]:
                # Raise an exception if asked to.
                self._handle_service_error(UPLOAD_CONFIG_ERROR_MESSAGE)

        except ConnectionError:
            # Raise an exception if asked to.
            self._handle_service_error(APP_SERVER_CONNECTION_ERROR_MESSAGE)
        except JSONDecodeError:
            # Raise an exception if asked to.
            self._handle_service_error(UPLOAD_CONFIG_ERROR_MESSAGE)

        return upload_output

    @Decorators.refresh_token_if_needed
    def upload_config_per_context_class(
        self, author, commit_message, context_class, config
    ):
        """
        A wrapper function for "Upload Config per Context Class" REST endpoint. view
        full documentation here:
        https://docs.monalabs.io/docs/upload-config-per-context-class-via-rest-api
        """
        app_server_response = self._app_server_request(
            "upload_config_for_context_class",
            data={
                "author": author,
                "commit_message": commit_message,
                "context_class": context_class,
                "config": config,
            },
        )
        if "response_data" not in app_server_response:
            self._handle_service_error(
                SERVICE_ERROR_MESSAGE + f" Service response: {app_server_response}"
            )

        return app_server_response

    @Decorators.refresh_token_if_needed
    def get_config(self):
        """
        :return: A json-serializable dict with the current defined configuration.
        """
        app_server_response = self._app_server_request("configs")
        try:
            return {
                self._user_id: app_server_response["response_data"][
                    "raw_configuration_data"
                ]
            }
        except KeyError:
            return self._handle_service_error(SERVICE_ERROR_MESSAGE)

    @Decorators.refresh_token_if_needed
    def get_suggested_config(self):
        """
        A wrapper function for "Retrieve Suggested Config" REST endpoint. view full
        documentation here:
        https://docs.monalabs.io/docs/retrieve-suggested-config-via-rest-api
        """
        app_server_response = self._app_server_request("get_new_config_fields")
        try:
            return app_server_response["response_data"]["suggested_config"]
        except KeyError:
            return self._handle_service_error(SERVICE_ERROR_MESSAGE)

    @Decorators.refresh_token_if_needed
    def get_config_history(self, number_of_revisions=UNPROVIDED_VALUE):
        """
        A wrapper function for "Retrieve Config History" REST endpoint. view full
        documentation here:
        https://docs.monalabs.io/docs/retrieve-config-history-via-rest-api
        """
        return self._app_server_request(
            "get_config_history", data={"number_of_revisions": number_of_revisions}
        )

    @cached(cache=TTLCache(maxsize=100, ttl=SAMPLING_FACTORS_MAX_AGE_SECONDS))
    def _update_sampling_factors_if_needed(self):
        """
        If the client was initiated with a sampling config name, check if the
        configuration was changed since the client vars were assigned, and if so, update
        them accordingly.
        """
        if not self._sampling_config_name:
            return

        # Refetch the updated config from the index.
        sampling_config = self.get_sampling_factors()[0]

        if self._latest_seen_sampling_config == sampling_config:
            return

        self._latest_seen_sampling_config = sampling_config
        default_from_index = sampling_config.get("default_factor")
        factors_map_from_index = sampling_config.get("factors_map")

        if (
            default_from_index is not None
            and default_from_index != self._default_sampling_rate
        ):
            logging.info(
                f"The default sampling factor was updated: {default_from_index}"
            )
            self._default_sampling_rate = default_from_index

        if (
            factors_map_from_index
            and factors_map_from_index != self._context_class_to_sampling_rate
        ):
            logging.info(
                f"The sampling factors map was updated: {factors_map_from_index}"
            )
            self._context_class_to_sampling_rate = factors_map_from_index

    @Decorators.refresh_token_if_needed
    def get_sampling_factors(self):
        """
        A wrapper function for "Get sampling factors" REST endpoint.
        The response will include a list of sampling config names, with their sampling
        map and default sampling factor, in the following format:
        [{
            "config_name": "Training",
            "factors_map": {
                "TEST_CONTEXT_CLASS": 0.5,
            },
            "default_factor": 0.1,
        }]
        config_name is a required field, factors_map and default_factor are optional.
        When the client is initiated with a config name, only the matching config
        details will be returned (if exists).
        """
        return self._app_server_request(
            "get_sampling_factors",
            data={"config_name": self._sampling_config_name},
        )["response_data"]

    @Decorators.refresh_token_if_needed
    def create_sampling_factor(self, config_name, sampling_factor, context_class=None):
        """
        A wrapper function for "Create sampling factor" REST endpoint.
        """
        response = self._app_server_request(
            "create_sampling_factor",
            data={
                "config_name": config_name,
                "sampling_factor": sampling_factor,
                "context_class": context_class,
            },
        )

        error_message = response["error_message"]

        return (
            f"Failed to create sampling factor: {error_message}"
            if error_message
            else "Sampling factor created successfully."
        )

    def validate_config(
        self,
        config,
        list_of_context_ids=UNPROVIDED_VALUE,
        latest_amount=UNPROVIDED_VALUE,
    ):
        """
        A wrapper function for "Validate Config" REST endpoint. view full documentation
        here: https://docs.monalabs.io/docs/validate-config-via-rest-api
        """
        return self._app_server_request(
            "validate_config",
            data={
                "config": config,
                "list_of_context_ids": list_of_context_ids,
                "latest_amount": latest_amount,
            },
        )

    @Decorators.refresh_token_if_needed
    def get_insights(
        self,
        context_class,
        min_segment_size,
        insight_types=UNPROVIDED_VALUE,
        metric_name=UNPROVIDED_VALUE,
        min_insight_score=UNPROVIDED_VALUE,
        time_range_seconds=UNPROVIDED_VALUE,
        first_discovered_on_range_seconds=UNPROVIDED_VALUE,
    ):
        """
        A wrapper function for "Retrieve Insights" REST endpoint. view full
        documentation here:
        https://docs.monalabs.io/docs/retrieve-insights-using-the-rest-api
        """
        return self._app_server_request(
            "insights",
            data={
                "context_class": context_class,
                "min_segment_size": min_segment_size,
                "insight_types": insight_types,
                "metric_name": metric_name,
                "min_insight_score": min_insight_score,
                "time_range_seconds": time_range_seconds,
                "first_discovered_on_range_seconds": first_discovered_on_range_seconds,
            },
        )

    @Decorators.refresh_token_if_needed
    def get_ingested_data_for_a_specific_segment(
        self,
        context_class,
        start_time,
        end_time,
        segment,
        excluded_segments=UNPROVIDED_VALUE,
    ):
        """
        A wrapper function for "Retrieve Ingested Data for a Specific Segment" REST
        endpoint. view full documentation here:
        https://docs.monalabs.io/docs/retrieve-ingested-data-for-a-specific-segment-via-rest-api
        """
        return self._app_server_request(
            "get_ingested_data",
            data={
                "context_class": context_class,
                "start_time": start_time,
                "end_time": end_time,
                "segment": segment,
                "excluded_segments": excluded_segments,
            },
        )

    @Decorators.refresh_token_if_needed
    def get_suggested_config_from_user_input(self, events):
        """
        A wrapper function for "Retrieve Suggested Config from User Input" REST
        endpoint. view full documentation here:
        https://docs.monalabs.io/docs/retrieve-suggested-config-from-user-input-via-rest-api
        """
        app_server_response = self._app_server_request(
            "suggest_new_config",
            data={"events": events},
        )
        if "response_data" not in app_server_response:
            self._handle_service_error(SERVICE_ERROR_MESSAGE)

        return app_server_response

    @Decorators.refresh_token_if_needed
    def get_aggregated_data_of_a_specific_segment(
        self,
        context_class,
        timestamp_from,
        timestamp_to,
        time_series_resolutions=UNPROVIDED_VALUE,
        with_histogram=UNPROVIDED_VALUE,
        time_zone=UNPROVIDED_VALUE,
        metrics=UNPROVIDED_VALUE,
        requested_segments=UNPROVIDED_VALUE,
        excluded_segments=UNPROVIDED_VALUE,
        baseline_segment=UNPROVIDED_VALUE,
    ):
        """
        A wrapper function for "Retrieve Aggregated Data of a Specific Segment" REST
        endpoint. view full documentation here:
        https://docs.monalabs.io/docs/retrieve-aggregated-data-of-a-specific-segment-via-rest-api
        """
        return self._app_server_request(
            "get_segment",
            data={
                "context_class": context_class,
                "timestamp_from": timestamp_from,
                "timestamp_to": timestamp_to,
                "time_series_resolutions": time_series_resolutions,
                "with_histogram": with_histogram,
                "time_zone": time_zone,
                "metrics": metrics,
                "requested_segments": requested_segments,
                "excluded_segments": excluded_segments,
                "baseline_segment": baseline_segment,
            },
        )

    @Decorators.refresh_token_if_needed
    def get_aggregated_stats_of_a_specific_segmentation(
        self,
        context_class,
        dimension,
        target_time_range,
        compared_time_range,
        metric_1_field,
        metric_2_field,
        metric_1_type,
        metric_2_type,
        min_segment_size,
        sort_function,
        baseline_segment=UNPROVIDED_VALUE,
        excluded_segments=UNPROVIDED_VALUE,
        target_segments_filter=UNPROVIDED_VALUE,
        compared_segments_filter=UNPROVIDED_VALUE,
    ):
        """
        A wrapper function for "Retrieve Aggregated Stats of Specific Segmentation" REST
        endpoint. view full documentation here:
        https://docs.monalabs.io/docs/retrieve-stats-of-specific-segmentation-via-rest-api
        """
        app_server_response = self._app_server_request(
            "stats",
            data={
                "context_class": context_class,
                "dimension": dimension,
                "target_time_range": target_time_range,
                "compared_time_range": compared_time_range,
                "metric_1_field": metric_1_field,
                "metric_2_field": metric_2_field,
                "metric_1_type": metric_1_type,
                "metric_2_type": metric_2_type,
                "min_segment_size": min_segment_size,
                "sort_function": sort_function,
                "baseline_segment": baseline_segment,
                "excluded_segments": excluded_segments,
                "target_segments_filter": target_segments_filter,
                "compared_segments_filter": compared_segments_filter,
            },
        )

        if "response_data" not in app_server_response:
            self._handle_service_error(
                f"{SERVICE_ERROR_MESSAGE} Server response: {app_server_response}"
            )

        return app_server_response

    def _handle_service_error(self, error_message):
        """
        Logs an error and raises MonaServiceException if RAISE_SERVICE_EXCEPTIONS is
        true, returns false otherwise.
        """
        error_message += self._get_unauthenticated_mode_error_message()

        self._logger.error(error_message)
        if self.raise_service_exceptions:
            raise MonaServiceException(error_message)
        return False

    def _get_unauthenticated_mode_error_message(self):
        """
        If should_use_authentication=False return an error message (suggesting
        should_use_authentication mode turned off might be the cause for the exception)
        and an empty string otherwise.
        """
        return (
            ""
            if self.should_use_authentication
            else UNAUTHENTICATED_ERROR_CHECK_MESSAGE
        )

    def _app_server_request(self, endpoint_name, data=None):
        """
        Send a request to Mona's app-server given endpoint with the given data (should
        be a dict with the endpoint requested fields).
        """
        try:
            app_server_response = requests.post(
                f"{self._app_server_url}/{endpoint_name}",
                headers=get_basic_auth_header(
                    self.api_key, self.should_use_authentication
                ),
                # Remove keys with UNPROVIDED_FIELD values to avoid overriding the
                # default value on the endpoint itself.
                json=remove_items_by_value(data, UNPROVIDED_VALUE) if data else {},
            )
            json_response = app_server_response.json()
            if not app_server_response.ok:
                return self._handle_service_error(SERVICE_ERROR_MESSAGE)

            return json_response

        except ConnectionError:
            return self._handle_service_error(APP_SERVER_CONNECTION_ERROR_MESSAGE)
        except JSONDecodeError:
            return self._handle_service_error(SERVICE_ERROR_MESSAGE)



# ----------------------------------------------------------------------------
#    Copyright 2021 MonaLabs.io
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
# ----------------------------------------------------------------------------
"""
This module contains all validation functions for client use.
"""
import json
import collections.abc

from .logger import get_logger
from .client_util import is_dict_contains_fields
from .client_exceptions import MonaExportException


def mona_messages_to_dicts_validation(
    events, raise_export_exceptions, log_failed_messages
):
    try:
        # Get all MonsSingleMessage as dicts.
        dict_events = [message.get_dict() for message in events]
    except TypeError:
        return handle_export_error(
            "export_batch must get an iterable of MonaSingleMessage.",
            raise_export_exceptions,
            events if log_failed_messages else None,
        )
    except AttributeError:
        return handle_export_error(
            "Messages exported to Mona must be MonaSingleMessage.",
            raise_export_exceptions,
            events if log_failed_messages else None,
        )

    # Validate that the batch is json serializable.
    if not _is_json_serializable(dict_events):
        return handle_export_error(
            "All fields in MonaSingleMessage must be JSON serializable.",
            raise_export_exceptions,
            events if log_failed_messages else None,
        )

    return dict_events


def _is_json_serializable(message):
    """
    Validates if the given message is a jsonable string.
    """
    try:
        json.dumps(message)
    except TypeError:
        return False

    return True


def validate_mona_single_message(message_event):
    # Check that message_event contains all required fields.
    required_fields = ("message", "contextClass")

    return is_dict_contains_fields(message_event, required_fields)


def update_mona_fields_names(message):
    """
    Changes names of fields that starts with "MONA_" to start with "MY_MONA_"
    """
    message_copy = dict(message)
    for key in message:
        if key.startswith("MONA_"):
            message_copy[f"MY_{key}"] = message_copy.pop(key)

    return message_copy


def validate_inner_message_type(message):
    """
    Validates the given input is a valid message (should be JSON serializable).
    """
    if not isinstance(message, collections.abc.Mapping):
        get_logger().error("Tried to send non-dict message to mona")
        return False

    return True


def handle_export_error(error_message, should_raise_exception, failed_message):
    """
    Logs an error and raises MonaExportException if RAISE_EXPORT_EXCEPTIONS is true,
    else returns false.
    """
    get_logger().error(error_message)
    if failed_message:
        get_logger().error(
            f"Failed to send the following to mona: {failed_message}"
        )
    if should_raise_exception:
        raise MonaExportException(error_message)
    return False



import asyncio
import json
from concurrent.futures import ThreadPoolExecutor

from mona_sdk import MonaSingleMessage
from mona_sdk.async_client import AsyncClient


async def main():
    client_id = "40c08ee6-f411-41b1-9752-80d98897aae6"
    secret_key = "5694fad0-e372-488e-89cf-22ed8693a5a2"
    user_id = "01971934-5acd-4888-9894-071e32fa867c"
    loop = asyncio.get_event_loop()

    with ThreadPoolExecutor(max_workers=1) as executor:

        async_client = AsyncClient(
            client_id, secret_key, user_id=user_id, event_loop=loop, executor=executor
        )

        print("Starting test ...")

        # is_active
        is_active = await async_client.is_active_async()
        print(f"is_active: {is_active}")

        # sampling factors
        sampling_factors = await async_client.get_sampling_factors_async()
        print(f"sampling_factors: {sampling_factors}")
        events = {
            "userId": user_id,
            "messages": [
                {
                    "arcClass": "arcClassName",
                    "contextId": "ID_123",
                    "message": {
                        "company-id": "79sg7723-0253438g3453",
                        "browser": "chrome",
                        "text_length": 50,
                        "top_tagged_brand": "adidas",
                    },
                },
                {
                    "arcClass": "arcClassName",
                    "contextId": "ID_123",
                    "message": {"coutry": "usa"},
                },
                {
                    "arcClass": "arcClassName",
                    "contextId": "ID_124",
                    "message": {
                        "company-id": "2435f32-02ve3538g5232",
                        "browser": "safari",
                        "text_length": 30,
                        "top_tagged_brand": "nike",
                    },
                },
            ],
        }
        suggested_config = (
            await async_client.get_suggested_config_from_user_input_async(events)
        )
        print(f"suggested_config: {suggested_config}")
        mona_messages = [
            MonaSingleMessage(
                message={"smadar_test": "1"},
                contextClass="SMADAR_TEST2",
                contextId="903c97aagfd6812b4c8e32fa",
            ),
            MonaSingleMessage(
                message={"smadar_test": "2"},
                contextClass="SMADAR_TEST1",
                contextId="903c97aagfd6812b48gj27df",
            ),
            MonaSingleMessage(
                message={
                    "smadar_test": "3"
                },
                contextClass="SMADAR_TEST1",
            ),
            MonaSingleMessage(
                message={
                    "smadar_test": "4"
                },
                contextClass="SMADAR_TEST_NOT_IN_SAMPLING_CONFIG",
            ),
            MonaSingleMessage(
                message={"smadar_test": "5"},
                contextClass="SMADAR_TEST_NOT_IN_SAMPLING_CONFIG",
                contextId="10jjf8e4gfd681204j38f7gb",
            ),
        ]
        # export (Single msg)
        export_res = await async_client.export_async(mona_messages[0])
        print(f"export res: {export_res}")
        # export batch
        export_batch_res = await async_client.export_batch_async(mona_messages)
        print(f"export batch res: {export_batch_res}")

        res = async_client.get_config()
        print(f"get_config res: {res}")

        config = {} # TODO: filll!!!!

        res = async_client.validate_config(
                config
        )
        print(f"validate_config res: {res}")

        with open("/home/smadar/workspace/mona-sdk/mona_sdk/mona_config_smadar_turorial.json") as fd:
            config = json.load(fd)
        if config["01971934-5acd-4888-9894-071e32fa867c"]["LOAN_APPLICATION_TUTORIAL"]["feature_9"]["bucket_size"] == 200:
            config["01971934-5acd-4888-9894-071e32fa867c"]["LOAN_APPLICATION_TUTORIAL"][
                "feature_9"]["bucket_size"] = 100


        # upload_config
        #res = async_client.upload_config(config, commit_message="smadar async test", author="smadar")

        # Upload_config_per_context_class
        #upload_config_per_context_class(self, author, commit_message, context_class, config)
        #get_config_history(self, number_of_revisions=UNPROVIDED_VALUE))
        #create_sampling_factor()
#
        #get_suggested_config_from_user_input(self, events)


    print("... Done!")


asyncio.run(main())



from types import FunctionType
from functools import wraps

def wrapper(method):
    @wraps(method)
    def wrapped(*args, **kwargs):
        print('{!r} executing'.format(method.__name__))
        return method(*args, **kwargs)
    return wrapped


class MetaClass(type):
    def __new__(meta, classname, bases, classDict):
        newClassDict = {}
        for attributeName, attribute in classDict.items():
            if isinstance(attribute, FunctionType):
                # replace it with a wrapped version
                attribute = wrapper(attribute)
            newClassDict[attributeName] = attribute
        return type.__new__(meta, classname, bases, newClassDict)


def with_metaclass(meta):
    """ Create an empty class with the supplied bases and metaclass. """
    return type.__new__(meta, "TempBaseClass", (object,), {})


if __name__ == '__main__':

    # Inherit metaclass from a dynamically-created base class.
    class MyClass(with_metaclass(MetaClass)):
        @staticmethod
        def a_static_method():
            pass

        @classmethod
        def a_class_method(cls):
            pass

        def a_method(self):
            pass

    instance = MyClass()
    instance.a_static_method()  # Not decorated.
    instance.a_class_method()   # Not decorated.
    instance.a_method()